from enum import Enum
from os import environ
from pathlib import Path
from sys import platform
from typing import Callable, Dict, List, Optional

import rich
from rich.text import Text


class Shell(Enum):
    BASH = "/bin/bash"
    ZSH = "/bin/zsh"

    @staticmethod
    def from_str(shell_str: Optional[str]) -> "Shell":
        if shell_str == Shell.BASH.value:
            return Shell.BASH
        if shell_str == Shell.ZSH.value:
            return Shell.ZSH
        raise KeyError(f"{shell_str} is not supported")


class Platform(Enum):
    LINUX = "linux"
    DARWIN = "darwin"


RC_FILE_NAME = {
    Shell.ZSH: ".zshrc",
    Shell.BASH: ".bashrc",
}


# Scripts were generated per https://click.palletsprojects.com/en/8.0.x/shell-completion/
COMPLETION_SCRIPTS = {
    Shell.BASH: """_layer_completion() {
    local IFS=$'
'
    COMPREPLY=( $( env COMP_WORDS="${COMP_WORDS[*]}" \
                   COMP_CWORD=$COMP_CWORD \
                   _LAYER_COMPLETE=complete $1 ) )
    return 0
}

_layer_completionetup() {
    local COMPLETION_OPTIONS=""
    local BASH_VERSION_ARR=(${BASH_VERSION//./ })
    # Only BASH version 4.4 and later have the nosort option.
    if [ ${BASH_VERSION_ARR[0]} -gt 4 ] || ([ ${BASH_VERSION_ARR[0]} -eq 4 ] && [ ${BASH_VERSION_ARR[1]} -ge 4 ]); then
        COMPLETION_OPTIONS="-o nosort"
    fi

    complete $COMPLETION_OPTIONS -F _layer_completion layer
}

_layer_completionetup;""",
    Shell.ZSH: """#compdef layer

_layer_completion() {
    local -a completions
    local -a completions_with_descriptions
    local -a response
    (( ! $+commands[layer] )) && return 1

    response=("${(@f)$(env COMP_WORDS="${words[*]}" COMP_CWORD=$((CURRENT-1)) _LAYER_COMPLETE=zsh_complete layer)}")

    for type key descr in ${response}; do
        if [[ "$type" == "plain" ]]; then
            if [[ "$descr" == "_" ]]; then
                completions+=("$key")
            else
                completions_with_descriptions+=("$key":"$descr")
            fi
        elif [[ "$type" == "dir" ]]; then
            _path_files -/
        elif [[ "$type" == "file" ]]; then
            _path_files -f
        fi
    done

    if [ -n "$completions_with_descriptions" ]; then
        _describe -V unsorted completions_with_descriptions -U
    fi

    if [ -n "$completions" ]; then
        compadd -U -V unsorted -a completions
    fi
}

compdef _layer_completion layer;""",
}

RC_FILE_CONTENT_GENERATOR: Dict[Shell, Callable[[Optional[Path]], List[str]]] = {
    Shell.BASH: lambda path: [f"source {path}"],
    # For below, see Section 20.2.1 at https://zsh.sourceforge.io/Doc/Release/Completion-System.html
    Shell.ZSH: lambda path: [  # NOTE: order of insertion for lines below matters
        "fpath+=~/.zfunc",  # add location for zsh autocomplete functions to search path
        "autoload -Uz compinit",  # Load the shell complete function
        "compinit",  # initialize completion
    ],
}


def _write_content_to_file_if_not_present(
    file_path: Path, content_to_write: List[str]
) -> None:
    file_path.parent.mkdir(parents=True, exist_ok=True)

    existing_lines = []
    if file_path.is_file():
        with open(file_path) as file:
            existing_lines = file.readlines()
            existing_lines = [line.rstrip() for line in existing_lines]

    existing_lines_set = set(existing_lines)
    for content_line in content_to_write:
        if content_line not in existing_lines_set:
            existing_lines.append(content_line)

    updated_content = "\n".join(existing_lines)
    file_path.write_text(updated_content)


def _create_file_and_write_content(file_path: Path, content_to_write: str) -> None:
    file_path.parent.mkdir(parents=True, exist_ok=True)
    file_path.write_text(content_to_write)


def _setup_zsh_completion() -> Path:
    # create completion script
    script_path = Path.home() / ".zfunc/_layer"
    script_content = COMPLETION_SCRIPTS[Shell.ZSH]
    _create_file_and_write_content(script_path, script_content)

    # Update zshrc to load completions
    file_path = Path.home() / RC_FILE_NAME[Shell.ZSH]
    content_to_write_generator = RC_FILE_CONTENT_GENERATOR[Shell.ZSH]
    content_to_write = content_to_write_generator(None)
    _write_content_to_file_if_not_present(file_path, content_to_write)
    return script_path


def _setup_bash_completion() -> Path:
    # create completion script
    completion_path = Path.home() / ".bash_completions/layer.sh"
    script_content = COMPLETION_SCRIPTS[Shell.BASH]
    _create_file_and_write_content(completion_path, script_content)

    # Update bashrc to load completions
    path_to_file = Path.home() / RC_FILE_NAME[Shell.BASH]
    content_to_write_generator = RC_FILE_CONTENT_GENERATOR[Shell.BASH]
    content_to_write = content_to_write_generator(completion_path)
    _write_content_to_file_if_not_present(path_to_file, content_to_write)
    return completion_path


def _print_completion_success_message(script_path: Path, rc_file_name: str) -> None:
    rich.print(
        Text.from_markup(":tada:").append(
            Text(
                f"Shell completion script installed under {script_path}; {rc_file_name} updated successfully",
                style="green",
            )
        )
    )
    rich.print(
        Text.from_markup(":electric_plug:").append(
            Text(
                "Restart your shell for shell completion to take effect",
                style="red bold",
            )
        ),
    )


def _check_completion_supported(shell_env_var: Optional[str]) -> bool:
    try:
        Platform[platform.upper()]
    except KeyError:
        rich.print(
            Text(
                f"Sorry, shell completion not yet supported for platform {platform}",
                style="red",
            )
        )
        return False

    try:
        Shell.from_str(shell_env_var)
    except KeyError:
        rich.print(
            Text(
                f"Sorry, shell completion not yet supported for shell {shell_env_var}",
                style="red",
            )
        )
        return False
    return True


def run_shell_completion_setup() -> None:
    shell_env_var = environ.get("SHELL")
    if _check_completion_supported(shell_env_var):
        if Shell.from_str(shell_env_var) == Shell.ZSH:
            script_path = _setup_zsh_completion()
            rc_file_name = RC_FILE_NAME[Shell.ZSH]
        else:
            script_path = _setup_bash_completion()
            rc_file_name = RC_FILE_NAME[Shell.BASH]

        _print_completion_success_message(script_path, rc_file_name)
