import asyncio
import logging
import sys
from pathlib import Path
from typing import List, Optional, Tuple

import click
import luddite  # type: ignore
import rich
from click_didyoumean import DYMGroup  # type: ignore
from click_help_colors import HelpColorsGroup  # type: ignore
from rich.console import Console
from rich.text import Text
from yarl import URL

from layer import __version__
from layer.async_utils import asyncio_run_in_thread
from layer.cli.click_validate import ValidUrl
from layer.cli.completion import run_shell_completion_setup
from layer.cli.prompt_integrations import (
    _prompt_big_query_integration,
    _prompt_snowflake_integration,
)
from layer.client import BigQueryDatasource, SnowflakeDatasource
from layer.common import LayerClient
from layer.config import DEFAULT_URL, Config
from layer.exceptions.exceptions import (
    IntegrationException,
    ProjectExecutionException,
    ProjectInitializationException,
    ProjectRunnerError,
)
from layer.main import ConfigManager, add_datasource, list_datasources
from layer.projects.entity import EntityType
from layer.projects.project import ProjectManager
from layer.projects.project_runner import ProjectRunner
from layer.projects.tracker.output import print_runs, watch_get_runs
from layer.projects.tracker.remote_execution_project_progress_tracker import (
    RemoteExecutionProjectProgressTracker,
)
from layer.user_logs import show_pipeline_run_logs


logger = logging.getLogger(__name__)


class AllGroups(HelpColorsGroup, DYMGroup):
    pass


def check_latest_version() -> None:
    latest_version = luddite.get_version_pypi("layer-sdk")
    current_version = __version__
    if current_version != latest_version:
        rich.print(
            Text.from_markup(":face_screaming_in_fear:").append(
                Text(
                    f" You are using the version {current_version} but the latest version is {latest_version}, please upgrade!",
                    style="bold red",
                )
            )
        )


@click.group(cls=AllGroups, help_headers_color="yellow", help_options_color="green")
@click.version_option(version=__version__, prog_name="Layer")
def cli() -> None:
    """Layer CLI to set up and run your declarative ML project."""
    check_latest_version()


@cli.command(short_help="Set up shell completion for Layer CLI commands")
def set_up_shell_completion() -> None:
    run_shell_completion_setup()


@cli.command(short_help="Log in to layer via web console.")
@click.argument("url", default=str(DEFAULT_URL), type=ValidUrl)
def login(url: str) -> None:
    """Log into layer via web console.

    URL is a link to Layer platform to log into.
    """
    asyncio.run(ConfigManager().login(URL(url)))


@cli.command(short_help="Log into layer via CLI.")
@click.argument("url", default=str(DEFAULT_URL), type=ValidUrl)
def login_headless(url: str) -> None:
    """Log into layer via CLI.

    URL is a link to Layer platform to log into.
    """
    asyncio.run(ConfigManager().login_headless(URL(url)))


@cli.command(short_help="Log out of your Layer account/organization.")
def logout() -> None:
    """Log out of your Layer account/organization."""
    asyncio.run(ConfigManager().logout())


@cli.group()
def integration() -> None:
    """Management of source integrations."""


@integration.command("add-snowflake", short_help="Configure Snowflake integration.")
@click.argument("name")
def cli_integration_add_snowflake(name: str) -> None:
    integration_add_snowflake(name=name, config_path=None)


def integration_add_snowflake(name: str, config_path: Optional[Path] = None) -> None:
    """Configure Snowflake integration.

    NAME a name for Snowflake source configuration.
    """
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    datasource = _prompt_snowflake_integration(name)
    add_datasource(datasource, config=config)
    print(f"Saved the Snowflake integration {name!r}")


@integration.command("add-big-query", short_help="Configure BigQuery integration.")
@click.argument("name")
def cli_integration_add_big_query(name: str) -> None:
    integration_add_big_query(name=name, config_path=None)


def integration_add_big_query(name: str, config_path: Optional[Path] = None) -> None:
    """Configure BigQuery integration.

    NAME a name for BigQuery source configuration.
    """
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    datasource = _prompt_big_query_integration(name)
    add_datasource(datasource, config=config)
    print(f"Saved the BigQuery integration {name!r}")


@integration.command("update", short_help="Update existing integration.")
@click.argument("name")
def cli_integration_update(name: str) -> None:
    integration_update(name=name, config_path=None)


def integration_update(name: str, config_path: Optional[Path] = None) -> None:
    """Update existing integration.

    NAME a name of an existing integration.
    """
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    with LayerClient(config.client, logger).init() as client:
        datasource = client.data_catalog.get_datasource_by_name(name)
        if isinstance(datasource, SnowflakeDatasource):
            print(f"Updating the Snowflake integration {name!r}")
            datasource = _prompt_snowflake_integration(name, id_=datasource.id)
        elif isinstance(datasource, BigQueryDatasource):
            print(f"Updating the BigQuery integration {name!r}")
            datasource = _prompt_big_query_integration(name, id_=datasource.id)
        else:
            raise IntegrationException(
                f"Unsupported integration type: {type(datasource).__name__}"
            )

        client.data_catalog.update_datasource(datasource)
        print(f"Saved the integration {name!r}")


@integration.command("ls", short_help="List existing integrations.")
def cli_integration_ls() -> None:
    integration_ls(config_path=None)


def integration_ls(config_path: Optional[Path] = None) -> None:
    """List existing integrations."""
    for datasource in list_datasources(config_path):
        print(datasource.name)


@cli.command("diagnose", short_help="Prints out information helpful for debugging.")
def diagnose() -> None:
    """Print out information that could be useful for debugging errors related to Layer."""
    ProjectManager().diagnose()


@cli.command("init", short_help="Initialize an empty project in the current folder.")
@click.argument(
    "directory",
    type=str,
    required=False,
    default=None,
)
def project_init(directory: Optional[str]) -> None:
    """
    Initialize an empty project into DIRECTORY. If none specified, the current folder will be used.
    Initializing into an existing directory is only allowed if the directory is empty.
    """
    if directory:
        ProjectManager().init(directory_path=Path(directory))
    else:
        ProjectManager().init()


@cli.command("clone")
@click.argument(
    "repo-url",
    type=str,
    default="https://github.com/layerml/examples",
)
@click.option(
    "-l",
    "--local-path",
    "local_path",
    default="",
    help="The local destination directory to save. Should be a relative path. It cannot be a file or a non-empty directory. The default is the name of the repo. If it does not exist, it will be automatically created.",
)
@click.option(
    "-b",
    "--branch",
    "branch",
    default="main",
    help="Which branch of the repository to pull from.",
)
def project_clone(repo_url: str, local_path: str, branch: str) -> None:
    """
    Clones a remote git repository to a local directory. The .git metadata folder is excluded.

    If REPO_URL is not specified, https://github.com/layerml/examples will be used.
    """
    ProjectManager().clone(repo_url, local_path, branch)


DEPENDENT_ENTITY_TYPE_NAMES = [EntityType.FEATURESET.value, EntityType.MODEL.value]
DEPENDENT_ENTITY_METAVAR = f"[[{'|'.join(DEPENDENT_ENTITY_TYPE_NAMES)}] NAME]"


@cli.command(
    "apply",
    short_help="Provision of Layer resources based on current project declaration.",
)
@click.argument(
    "type_and_name",
    type=(click.Choice(DEPENDENT_ENTITY_TYPE_NAMES + [""]), str),
    metavar=DEPENDENT_ENTITY_METAVAR,
    required=False,
    default=("", ""),
)
@click.option(
    "-d/-D",
    "--dependencies/--no-dependencies",
    "keep_dependencies",
    is_flag=True,
    default=False,
    help="Include dependencies if NAME is specified",
)
def cli_project_apply(
    type_and_name: Tuple[str, str],
    keep_dependencies: bool,
) -> None:
    project_apply(
        type_and_name=type_and_name,
        keep_dependencies=keep_dependencies,
        config_path=None,
    )


def project_apply(
    type_and_name: Tuple[str, str],
    keep_dependencies: bool,
    config_path: Optional[Path] = None,
) -> None:
    """Provision of Layer resources based on current project declaration."""
    type_, name = type_and_name
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    _print_version(config)
    project = ProjectManager().load()
    if name:
        project = project.drop_independent_entities(
            EntityType(type_), name, keep_dependencies=keep_dependencies
        )
    ProjectRunner(
        config, project_progress_tracker_factory=RemoteExecutionProjectProgressTracker
    ).apply(project)


@cli.command(
    "start",
    short_help="Calls 'apply' and then "
    "creates the defined featureset and trains the model.",
)
@click.argument(
    "type_and_name",
    type=(click.Choice(DEPENDENT_ENTITY_TYPE_NAMES + [""]), str),
    metavar=DEPENDENT_ENTITY_METAVAR,
    required=False,
    default=("", ""),
)
@click.option(
    "-d/-D",
    "--dependencies/--no-dependencies",
    "keep_dependencies",
    is_flag=True,
    default=False,
    help="Include dependencies of NAME as part of the build. Has no effect without NAME.",
)
def cli_project_start(
    type_and_name: Tuple[str, str],
    keep_dependencies: bool,
) -> None:
    project_start(
        type_and_name=type_and_name,
        keep_dependencies=keep_dependencies,
        config_path=None,
    )


def project_start(
    type_and_name: Tuple[str, str],
    keep_dependencies: bool,
    config_path: Optional[Path] = None,
) -> None:
    """Creates the defined featureset and trains the model.
    type_and_name argument can be used to start only specific featureset or model for instance
    one can train a specific model by calling start with model name as argument `layer start model <model_name>`"""
    type_, name = type_and_name
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    _print_version(config)
    project = ProjectManager().load()
    if name:
        project = project.drop_independent_entities(
            EntityType(type_), name, keep_dependencies=keep_dependencies
        )
    ProjectRunner(
        config, project_progress_tracker_factory=RemoteExecutionProjectProgressTracker
    ).run(project)


@cli.command(
    "runs",
    short_help="Lists a single or multiple running pipelines",
)
@click.argument("pipeline_run_id", default="")
@click.option(
    "-w/-W",
    "--watch/--no-watch",
    "watch",
    is_flag=True,
    default=False,
    help="Watch the specified pipeline runs",
)
@click.option(
    "-a/-A",
    "--all/--active",
    "list_all",
    is_flag=True,
    default=False,
    help="Display all pipelines when this option is set. Otherwise display only active pipelines",
)
def cli_list_runs(
    pipeline_run_id: str,
    watch: bool,
    list_all: bool,
) -> None:
    project = ProjectManager().load()
    list_runs(
        pipeline_run_id=pipeline_run_id,
        watch=watch,
        list_all=list_all,
        project_name=project.name,
        config_path=None,
    )


def list_runs(
    pipeline_run_id: str,
    watch: bool,
    list_all: bool,
    project_name: str,
    config_path: Optional[Path] = None,
) -> None:
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    console = Console()
    with LayerClient(config.client, logger).init() as client:
        runs, get_runs_func = client.flow_manager.get_runs(
            pipeline_run_id, list_all, project_name
        )
        print_runs(runs)
        if watch:
            watch_get_runs(console, get_runs_func)


@cli.command(
    "kill",
    help="Terminates a pipeline run for the provided run id",
    short_help="Terminates a pipeline run for the provided run id",
)
@click.argument(
    "run-id",
    type=str,
    required=True,
)
def cli_project_kill_run(
    run_id: str,
) -> None:
    project_kill_run(run_id=run_id, config_path=None)


def project_kill_run(run_id: str, config_path: Optional[Path] = None) -> None:
    config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
    _print_version(config)
    ProjectRunner(
        config, project_progress_tracker_factory=RemoteExecutionProjectProgressTracker
    ).terminate_run(run_id)


def _print_version(config: Config) -> None:
    rich.print(
        Text.assemble(
            (f"Layer {__version__}", "bold"),
            " using ",
            (f"{config.url}", "underline"),
        )
    )


@cli.command(
    "logs",
    short_help="Shows the logs produced by running a pipeline",
)
@click.argument(
    "pipeline_run_id",
    type=str,
    required=True,
)
@click.option(
    "-f",
    "--follow",
    "follow",
    is_flag=True,
    default=False,
    help="Specify if the logs should be continuously followed",
)
def cli_user_logs(
    pipeline_run_id: str,
    follow: bool,
) -> None:
    user_logs(pipeline_run_id=pipeline_run_id, follow=follow, config_path=None)


def user_logs(
    pipeline_run_id: str, follow: bool, config_path: Optional[Path] = None
) -> None:
    try:
        config = asyncio_run_in_thread(ConfigManager(config_path).refresh())
        with LayerClient(config.client, logger).init() as client:
            show_pipeline_run_logs(client, pipeline_run_id, follow)
    except KeyboardInterrupt:
        sys.exit()


def main(args: Optional[List[str]] = None) -> None:
    try:
        cli.main(args=args)
    except ProjectInitializationException as e:
        rich.print(
            Text.from_markup(":x:").append(
                Text(
                    " Project initialization error.",
                    style="bold red",
                )
            )
        )
        rich.print(
            Text.from_markup(":mag_right:").append(
                Text.from_markup(f" {e.error_msg_rich}")
            )
        )
        rich.print(
            Text.from_markup(":bulb:").append(
                Text.from_markup(
                    f" {e.suggestion_rich}",
                )
            )
        )
        sys.exit(1)
    except ProjectExecutionException as e:
        rich.print(
            Text.from_markup(":x:").append(
                Text(
                    " Project execution error.",
                    style="bold red",
                )
            )
        )
        rich.print(
            Text.from_markup(":mag_right:").append(
                Text.from_markup(f" {e.error_msg_rich}")
            )
        )
        rich.print(
            Text.from_markup(":bulb:").append(
                Text.from_markup(
                    f" {e.suggestion_rich}",
                )
            )
        )
        sys.exit(1)
    except ProjectRunnerError:
        sys.exit(1)
    except Exception as exc:
        print(str(exc) or type(exc).__name__)
        sys.exit(1)
