import uuid
from pathlib import Path
from typing import Any, MutableMapping, Optional

from prompt_toolkit import prompt
from prompt_toolkit.validation import Validator

from layer.client import BigQueryDatasource, Datasource, SnowflakeDatasource


def _is_not_empty(value: str) -> bool:
    return bool(value)


def _is_file(value: str) -> bool:
    path = Path(value)
    return path.exists() and path.is_file()


_STR_VALIDATOR = Validator.from_callable(_is_not_empty)
_FILE_VALIDATOR = Validator.from_callable(_is_file)


def _prompt_snowflake_integration(
    name: str, *, id_: Optional[uuid.UUID] = None
) -> Datasource:
    kwargs: MutableMapping[str, Any] = {"name": name}
    if id_:
        kwargs["id"] = id_
    validator = _STR_VALIDATOR
    return SnowflakeDatasource(
        url=prompt("URL: ", validator=validator, validate_while_typing=False),
        user=prompt("User: ", validator=validator, validate_while_typing=False),
        password=prompt(
            "Password: ",
            validator=validator,
            is_password=True,
            validate_while_typing=False,
        ),
        database=prompt("Database: ", validator=validator, validate_while_typing=False),
        schema=prompt("Schema: ", validator=validator, validate_while_typing=False),
        role=prompt("Role [empty for none]: ", validator=None),
        warehouse=prompt(
            "Warehouse: ", validator=validator, validate_while_typing=False
        ),
        **kwargs,
    )


def _prompt_big_query_integration(
    name: str, *, id_: Optional[uuid.UUID] = None
) -> Datasource:
    kwargs: MutableMapping[str, Any] = {"name": name}
    if id_:
        kwargs["id"] = id_

    str_validator = _STR_VALIDATOR
    path_validator = _FILE_VALIDATOR
    credentials_path = Path(
        prompt(
            "Service Account Key Path: ",
            validator=path_validator,
            validate_while_typing=False,
        )
    )
    return BigQueryDatasource(
        credentials=credentials_path.read_text(),
        project_id=prompt(
            "Project ID: ", validator=str_validator, validate_while_typing=False
        ),
        dataset_id=prompt(
            "Dataset ID: ", validator=str_validator, validate_while_typing=False
        ),
        **kwargs,
    )
